# --------------------------------------------------------------------------
###
#
# Copyright (C) Ensim Corporation 2000, 2001   All Rights Reserved.
#
# This software is furnished under a license and may be used and copied
# only  in  accordance  with  the  terms  of such  license and with the
# inclusion of the above copyright notice. This software or any other
# copies thereof may not be provided or otherwise made available to any
# other person. No title to and ownership of the software is hereby
# transferred.
#
# The information in this software is subject to change without notice
# and  should  not be  construed  as  a commitment by Ensim Corporation.
# Ensim assumes no responsibility for the use or  reliability  of its
# software on equipment which is not supplied by Ensim.
#
# $Id: DnsUpdate.pm,v 1.1.1.1 2008/03/01 00:21:34 aarefyev Exp $
# --------------------------------------------------------------------------

package DnsUpdate;

use strict;
use IPC::Open3;
use IO::Handle;
use IO::Select;

use DnsUpdateResult;
use DnsCommon;
use Zone_TSIG qw/getZoneTSIGPrivateKeyFile/;
use dns qw/zoneExists zoneType zoneMasters zoneForwarders/;
use dnsX qw/getZoneUpdateKey getHostUpdateKey getDefaultLocalUpdateKey/;

use constant DEBUG => 0;

sub new () {
    my $class = shift;
    my $self  = {};

    bless $self, $class;
    $self->_init (@_);
    
    return $self;
}

sub pushPrereq {
    my $self = shift;
    my $op   = shift;
    my $val = shift;

    if ($op eq "nxdomain") {
        push @{$self->{prereq}}, "nxdomain $val";
    } elsif ($op eq "yxdomain") {
        push @{$self->{prereq}}, "yxdomain $val";
    } elsif ($op eq "nxrrset") {
        # we support only IN type
        push @{$self->{prereq}}, "nxrrset $val";
    } elsif ($op eq "yxrrset") {
        my $data = shift;
        $data or ($data = "");
        push @{$self->{prereq}}, "yxrrset $val $data";
    }
}

sub pushUpdate {
    my $self = shift;
    my $op   = shift;

    # class is assumed to be IN
    my $name;
    my $ttl;
    my $type;
    my $data;

    if ($op eq "delete") {
        $name = shift;
        ($type = shift) or ($type = "");
        ($data = shift) or ($data = "");
        
        push @{$self->{update}}, "delete $name $type $data";
    } elsif ($op eq "add") {
        $name = shift;
        $ttl  = shift;
        $type = shift;
        $data = shift;

        push @{$self->{update}}, "add $name $ttl $type $data";
    }
}
sub send {
    my $self = shift;
    
    my ($cin, $cout, $cerr) = (IO::Handle->new, IO::Handle->new, 
                             IO::Handle->new);
    my $pid;
    my $childerr=0;
    
    my $cmd = $self->_getCmd();

    # We will no longer use the files in /etc/bind/.TSIG/* in order to pass to nsupdate
    # We will pass on the key over the nsupdate channel itself. Much cleaner than the old approach,
    # especially since it is not always assumed that the TSIG key is the one that we are using 

    # my $cmdArgs = $self->_getNsupdateArgs();
    # $pid = open3($cin, $cout, $cerr, "$cmd $cmdArgs");

    $pid = open3 ($cin, $cout, $cerr, "$cmd -d");       # -d for debugging

    $SIG{CHLD} = sub {
        if (waitpid ($pid, 0) > 0) {
            $childerr = ($? >> 8);
        }
    };

    # use local as server by default
    print $cin "server $self->{server}\n";
    if (DEBUG) {
        print STDERR ">server $self->{server}\n";
    }

    if ($self->{zone}) {
        print $cin "zone $self->{zone}\n";
        if (DEBUG) {
            print STDERR ">zone $self->{zone}\n";
        }
    }

    # supply the key name and the secret if we know it
    if ($self->{keyname} and $self->{secret}) {
        print $cin "key $self->{keyname} $self->{secret}\n";
        if (DEBUG) {
            print STDERR ">key $self->{keyname} $self->{secret}\n";
        }
    }

    foreach my $prereq (@{$self->{prereq}}) {
        print $cin "prereq $prereq\n";
        if (DEBUG) {
            print STDERR ">prereq $prereq\n";
        }
    }
    
    foreach my $update (@{$self->{update}}) {
        print $cin "update $update\n";
        if (DEBUG) {
            print STDERR ">update $update\n";
        }
    }

    print $cin "\n";

    $cin->close();

    my $selector = IO::Select->new();
    $selector->add($cerr, $cout);
    
    my @ready;
    my $buf;
    my @message = ();
    while (@ready = $selector->can_read) {
        foreach my $fh (@ready) {
            if ($fh->fileno == $cerr->fileno) {
                $buf = <$fh>;
                if (DEBUG) {
                    print STDERR $buf;
                }

		if ($buf) {
		    chomp $buf;
		    push @message, $buf;
		}
            } elsif ($fh->fileno == $cout->fileno) {
                # discard stdout
                $buf = <$fh>;
            }

            $selector->remove($fh) if $fh->eof;
        }
    }

    if (DEBUG) {
	print "--END OF output--\n";
    }

    $cout->close;
    $cerr->close;
    
    # handle the result
    if ($childerr == 1) {
	# fatal error, last line == error
	$self->{errorstring} = pop @message;
	chomp $self->{errorstring};

    } else {
	###
	# we would like to look for:
	#  - "Reply from update query"
	#  - "Found zone name:..."
	#  - "The master is:..."
	#  - ";...."
	###

	my $zone;
	my $master;
	my @headerMsg = ();

	# precompile the patterns
	my $zonePattern   = qr/^Found zone name: (.*)$/;
	my $masterPattern = qr/^The master is: (.*)$/;
	my $replyPattern  = qr/^Reply from update query:\w*$/;
	my $errPattern    = qr/^; (.*)$/;
	    
	for (my $i=0; $i<scalar(@message); $i++) { 
	    if ($message[$i] =~ /$zonePattern/) {
		$zone = $1;
	    } elsif ($message[$i] =~ /$masterPattern/) {
		$master = $1;
	    } elsif ($message[$i] =~ /$replyPattern/) {
		push @headerMsg, $message[++$i];
		push @headerMsg, $message[++$i];
		
		last;
	    } elsif ($message[$i] =~ /$errPattern/) {
		# error case !!
		$self->{errorstring} = $1;
		chomp $self->{errorstring};

		return;
	    }
	}

	# if we have set zone in the transaction
	# there will be no master, and zone from stderr
	if ($self->{zone}) {
	    $zone = $self->{zone};
	    $master = "";
	}
	
	return new DnsUpdateResult $zone, $master, \@headerMsg;
    }

    return;
}

sub errorstring {
    my $self = shift;
    return $self->{errorstring};
}

# This method would fill out a DnsUpdate Object
# with the right information to perform SOA update
# 
# Returns:
#  - false if nothing has to be done
#  - true  if update has to be performed
#
# Exception:
#  - SOA lookup failure
sub updateSOA {
    my $class = shift;
    my $updater = shift;
    my $zone = shift;
    my $newValues = shift;
    my $rnameAutoUpdate = shift; # optional, default false

    # get the current SOA record
    my $res = DnsCommon::getResolver();
    my $query = $res->query($zone, "SOA");

    my $cursoa;
    if ($query) {
	$cursoa = ($query->answer)[0];
    } else {
	die $res->errorstring . "\n";
    }
    

    # convert the current RData into string and assigning
    # existing value
    my @curRData;
    my @newRData;
    
    # increment the serial number, unless the caller provided one
    if(!$newValues->{serial}) {
        $newValues->{serial} = DnsCommon::incSerial ($cursoa->{serial});
    }

    my $updateRequired = -1;

    # handle sync of rname to mname
    if ($rnameAutoUpdate &&  (!$newValues->{rname}) &&
	($newValues->{mname}) && $cursoa->rname) {

	# see if the mail domain is the same as mname
	my $adminMailDomain = $cursoa->rname;
	$adminMailDomain =~ s/^([^\.]*)\.?//;
	my $adminMailUser = "$1";

	if (lc($adminMailDomain) eq lc($cursoa->mname)) {
	    # we need to update the rname mail domain
	    $newValues->{rname} = sprintf ("%s.%s", $adminMailUser, 
					   $newValues->{mname});
	}
    }
    
    # need to add '.' to the end of email and master
    map {
	push @curRData, $cursoa->$_ . '.';
	push @newRData, ($newValues->{$_} || $cursoa->$_ . '.');
	
	if ($newValues->{$_} &&
	    ($newValues->{$_} ne ($cursoa->$_ . '.'))) {
	    $updateRequired++;
	}
    } ("mname", "rname");
    
    
    map {
	push @curRData, $cursoa->$_;
	push @newRData, ($newValues->{$_} || $cursoa->$_);
	
	if ($newValues->{$_} &&
	    ($newValues->{$_} ne $cursoa->$_)) {
	    $updateRequired++;
	}
    } ("serial", "refresh", "retry", "expire", "minimum");

    
    if (!$updateRequired) {
	return "";
    }


#    $updater->pushPrereq ("yxrrset", join (' ', '@ IN SOA ', @curRData));
#    $updater->pushUpdate ("add", '@',
#			DnsCommon::getDefaultRRTTL(), "SOA",
#			  join ' ', @newRData);

     $updater->pushPrereq ("yxrrset", join (' ', "$zone IN SOA ", @curRData));
     $updater->pushUpdate ("add", "$zone",
                        DnsCommon::getDefaultRRTTL(), "SOA",
                          join ' ', @newRData); 
    return 1;
}

# It is now required that the DNS zone updater class be allowed to accept two 
# arguments instead of one. 
# Earlier, the DnsUpdate constructor only ever expected the zone name to 
# be passed to it. However, now, we may also wish to pass a host name argument to 
# the DnsUpdate constructor. 
# This would be useful only if the zone that we wish to update, does not reside
# locally
sub _init { 
    my $self = shift;
    my $zone = shift;
    my $server = shift;

    my @pre = ();
    $self->{prereq} = \@pre;

    my @update = ();
    $self->{update} = \@update;

    # Indicate whether we know how to connect to the correct server
    $self->{can_connect} = 0;

    if ($zone && zoneExists ($zone)) {
        # The zone name has been specified AND it exists locally
        $self->{zone} = $zone;

        # The zone exists locally
        if ( zoneType ($zone) eq 'master' ) {
            # and is a master
            $self->{server} = '127.0.0.1';
            # get the TSIG keyname and secret which we will use to communicate with the server to update the zone
            ($self->{keyname}, $self->{secret}) = getZoneUpdateKey ($zone);
            $self->{can_connect} = 1;
        } elsif ( zoneType ($zone) eq 'slave') {
            # and is a slave
            $self->{server} = @{zoneMasters ($zone)}[0];
            ($self->{keyname}, $self->{secret}, $self->{can_connect}) = getHostUpdateKey ($self->{server});
        } elsif ( zoneType ($zone) eq 'forward') {
            # and is a forward zone
            my @forwarders = @{zoneForwarders ($zone)};
            # the first among the forwarders is interpreted as the master, the rest are slaves
            if ($#forwarders > -1) { $self->{server} = $forwarders[0]; }
            else { $self->{server} = '127.0.0.1'; };
            ($self->{keyname}, $self->{secret}, $self->{can_connect}) = getHostUpdateKey ($self->{server});
        }
        else {
            #warn "Defaulting to local server, although zone exists and is neither master nor slave\n";
            $self->{server} = '127.0.0.1';
            ($self->{keyname}, $self->{secret}) = getDefaultLocalUpdateKey ();
            $self->{can_connect} = 1;
        }
    }
    else {
        # Zone has not been specified, or it has been specified but the zone does not exist

        if ($server) {
            $self->{server} = $server;  # XXX: or should we use its IP address ?
            ($self->{keyname}, $self->{secret}, $self->{can_connect}) = getHostUpdateKey ($server);
        } else {
            #warn "Defaulting to local server, since no zone specified and no server specified\n";
            $self->{server} = '127.0.0.1';
            ($self->{keyname}, $self->{secret}) = getDefaultLocalUpdateKey ();
            $self->{can_connect} = 1;
        }
    }
}

sub _getNsupdateArgs($) {
    my $self = shift;
    my $TSIG_keyFile = getZoneTSIGPrivateKeyFile($self->{zone});

    return "-k $TSIG_keyFile -d";
}

sub _getCmd($) {
    my $self = shift;

    return "/usr/bin/nsupdate";
}

1;




