# dnsX -- A package containing new DNS features added to Ensim Pro X
#

# $Id: dnsX.pm,v 1.1.1.1 2008/03/01 00:21:33 aarefyev Exp $

package dnsX;
use strict;
require Exporter;

our @ISA = qw /Exporter/;
our @EXPORT_OK = qw /zoneExists zoneType zoneMasters isLocal getZoneUpdateKey getHostUpdateKey getDefaultLocalUpdateKey lookupHost getLocalHostName zoneForwarders getLocalIP/;

use lib ($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter")."/bind";
use lib ($ENV{'OCW_SVCPATH'} || "/usr/lib/opcenter")."/cmdline_common";

use ZoneParser;
use Config::IniFiles;
use Sys::Hostname qw (hostname);
use File::Basename;
use LogConfFile;

use constant BIND_CONF => "/etc/named.conf";
use constant LOCALHOST_FILE => "/etc/virtualhosting/localnameserver";
use constant CONF_LOG_DIR => "/var/log/conf";

# returns a reference to a hash of zone information, keyed by zone name
sub getZones () {
    my $parser = new ZoneParser;
    $parser->parse_file(BIND_CONF);
    
    return $parser->zones;
}

# Check to see if a zone exists. returns 0/1
sub zoneExists ($) {
    my $z = shift;
    my $zones = getZones ();
    return exists ($zones->{$z}) ? 1 : 0;
}

# Check to see if a zone is a master or a slave on the local server.
# Call this function only if you know that the zone exists locally
sub zoneType ($) {
    my $z = shift;
    my $zones = getZones ();
    return $zones->{$z}->{type};
}

# Returns a reference to an array containing the list of the master servers of a zone
sub zoneMasters ($) {
    my $z = shift;
    my $zones = getZones ();
    die "Zone $z is not a slave (hence, no masters)\n" unless zoneExists($z) and zoneType ($z) eq 'slave';
    return $zones->{$z}->{masters};
}

# Returns a reference to an array containing the list of forwarders for a forward-type zone
sub zoneForwarders ($) {
    my $z = shift;
    my $zones = getZones ();
    die "Zone $z is not a forward zone (hence, no forwarders)\n" unless zoneExists($z) and zoneType ($z) eq 'forward';
    my @forwarders = ();
    if ((exists $zones->{$z}->{forwarders}) and (ref ($zones->{$z}->{forwarders}) eq 'ARRAY')) {
        @forwarders = @{$zones->{$z}->{forwarders}};
    }
    return \@forwarders;
}

# Returns 0/1 depending on whether the argument passed to the function identifies the local node
sub isLocal ($) {
    my $h = shift;              # The host name we have to look up
    my @l = getLocalHostName ();    # The name of the local system

    my ($hname, $haliases, $haddrtype, $hlength, @haddrs);
    ($hname, $haliases, $haddrtype, $hlength, @haddrs) = gethostbyname ($h);

    my @hipaddrs = ();
    foreach (@haddrs) {
        push (@hipaddrs, join (".", unpack ('C4', $_)));
    }

    my $found = 0;

    for my $lh (@l) {
        #print "\nlocal host name : ",$lh,"\n";
        my ($lname, $laliases, $laddrtype, $llength, @laddrs);
        ($lname, $laliases, $laddrtype, $llength, @laddrs) = gethostbyname ($lh);

        my @lipaddrs = ();
        foreach (@laddrs) {
            push (@lipaddrs, join (".", unpack ('C4', $_)));
        }
        #print "local host names ipaddrs : ", @lipaddrs,"\n";

        for my $hipaddr (@hipaddrs) {
            #print "host ipaddr : ", $hipaddr,"\n";
            if (grep (/^$hipaddr$/, @lipaddrs)) {
                #print "found one";
                $found = 1;
            }
        }
    }

    #print "found = ",$found,"\n";
    return $found;
}

# returns the first IP address matching a given a host name, or returns the hostname itself
# if the name does not resolve
sub lookupHost ($) {
    my $h = shift;
    my ($name,$aliases,$addrtype,$length,@addrs) = gethostbyname ($h);
    if ($#addrs == -1) {
        return $h;
    } else {
        return join ".", map { unpack ('C4',$_) } $addrs[0];
    }
}

# given a zone name, gets the key name and secret required to update that zone.
# Returns keyname:secret, useful for passing to nsupdate with the -y option
sub getZoneUpdateKey ($) {
    my $zp = new ZoneParser;
    $zp->parse_file (BIND_CONF);
    my $zone = shift;
    my $keyname = "";
    my $secret = "";
    my $found = 0;

    # We will only search locally available zones for the allow-update directive, because
    # slave zones are not allowed 
    if (exists $zp->{zones}->{$zone} and $zp->{zones}->{$zone}->{type} eq 'master') {
        # The zone was found in the conf file. Now search for the key
        if (exists $zp->{zones}->{$zone}->{'allow-update'}) {
            my $anticipate_key = 0;

            for my $v (@{$zp->{zones}->{$zone}->{'allow-update'}[0]}) {
                if ($anticipate_key == 1) { $found = 1; $keyname = "$v"; last; }
                if ("$v" eq "key") { $anticipate_key = 1; }
            }

            if ($keyname) {
                if (exists $zp->{_keys} && exists $zp->{_keys}->{$keyname}) {
                    $secret = $zp->{_keys}->{$keyname};
                }
            }
        }
    }

    return $keyname, $secret;
}

# given a host name, attempt to look up a locally maintained INI file and find out the key
# and secret from there
sub getHostUpdateKey ($) {
    my $h = shift;
    my $i = lookupHost ($h);
    my $keyname = "";
    my $secret = "";
    my $canupdate = 0;

    my $confparser = new Config::IniFiles (-file => '/etc/virtualhosting/masterconf');
    if (grep /^$i$/, @{$confparser->{sects}}) {
        $keyname = $confparser->{v}->{$i}->{key} || "";
        $secret = $confparser->{v}->{$i}->{secret} || "";
        $canupdate = $confparser->{v}->{$i}->{update_zones} || 0;
    } else {
        warn "Cannot find $i in External DNS Server Map";
    }

    return $keyname, $secret, $canupdate;
}

# returns the keyname and secret that can be used to update the local EPL server
sub getDefaultLocalUpdateKey () {
    my $zp = new ZoneParser;
    my ($keyname, $secret);

    $zp->parse_file (BIND_CONF);

    if (exists $zp->{_keys} and exists $zp->{_keys}->{'wp_default_key.'}) {
        $keyname = 'wp_default_key.';   # quite redundant
        $secret = $zp->{_keys}->{'wp_default_key.'};
    }

    return $keyname,$secret;
}

# Returns the FQDN of the local node, after doubly verifying it against a DNS resolver, if possible
sub getLocalHostName() {
    my @localhostnames = ();

    # if file is not present create it.
    if (! -e LOCALHOST_FILE) {
        open  TOUCH_FILE, ">" . LOCALHOST_FILE;
        close TOUCH_FILE;
    }

    if (-r LOCALHOST_FILE) {
      my $buf;

      my $localhostData = new IO::File LOCALHOST_FILE, "r";
      while ($buf = <$localhostData>) {
        chomp $buf;
        $_ = $buf;
        my ($host) = split(' ');
        push @localhostnames, $host;
      }
      # if file was empty add entry for hostname from /etc/hosts.
      my $size = scalar @localhostnames;
      if ($size == 0) {
        my $fileDir = dirname LOCALHOST_FILE;
        my $filename = basename LOCALHOST_FILE;

        #my $lhData = new IO::File LOCALHOST_FILE, "w";
        my $lhData = LogConfFile::open_conf ($fileDir, CONF_LOG_DIR, $filename, "w", 1);

        my $hostIP;
        my $hostnameIP;
        my $hostname = hostname;
        $hostnameIP = gethostbyname($hostname);
        $hostIP = join(".",(unpack('C4',$hostnameIP)));
        print $lhData "$hostname $hostIP\n";

        #$lhData->close;
        LogConfFile::close_conf($lhData, $fileDir, CONF_LOG_DIR, $filename);
        push(@localhostnames,$hostname);
      }
      $localhostData->close;
    }
    return @localhostnames
}

sub getLocalIP() {
    my @localIPaddr = ();

    # if file is not present create it.
    if (! -e LOCALHOST_FILE) {
        open  TOUCH_FILE, ">" . LOCALHOST_FILE;
        close TOUCH_FILE;
    }

    if (-r LOCALHOST_FILE) {
      my $buf;

      my $localIPData = new IO::File LOCALHOST_FILE, "r";
      while ($buf = <$localIPData>) {
        chomp $buf;
        $_ = $buf;
        my ($host,$IP) = split(' ');
        if ($IP) {
            push @localIPaddr, $IP;
        }
      }
      $localIPData->close;
    }
    return @localIPaddr;
}

1;
